from ._http import (
    _get_channel_about,
    _get_channel_live_data,
    _get_old_streams,
    _get_uploads_data,
    _get_channel_playlists,
    _get_video_count,
    _get_upcoming_videos
)
from typing import List, Optional, Dict, Any
from .live import Live
from .video import Video
from ._threads import _Thread
from urllib.parse import unquote
from .videobulk import _VideoBulk
from .utils import dup_filter
from .upcoming import Upcoming
from .playlistbulk import _PlaylistBulk
from ._rgxs import _ChannelPatterns as rgx


class Channel:

    __HEAD = 'https://www.youtube.com/channel/'
    __CUSTOM = 'https://www.youtube.com/c/'

    def __init__(self, channel_id: str):

        if channel_id is None:
            raise ValueError('channel id or url cannot be None')

        if channel_id.startswith('UC'):
            self._url = self.__HEAD + channel_id
        elif channel_id.startswith('http'):
            self._url = channel_id
        else:
            self._url = self.__CUSTOM + channel_id
        self.__given_id = channel_id

    def __repr__(self):
        if self.id:
            return f'<Channel {self.url}>'
        return '<Invalid Channel Object>'

    @property
    def __raw_about(self) -> str:
        return _get_channel_about(self._url)

    @property
    def name(self) -> Optional[str]:

        name = rgx.name.findall(self.__raw_about)
        return name[0] if name else None

    @property
    def valid(self) -> bool:

        source = _src(self._url)
        return True if source else False

    @property
    def url(self) -> str:
        return self.__HEAD + self.id

    @property
    def id(self) -> str:

        channel_id = rgx.id.findall(self.__raw_about)
        return channel_id[0] if channel_id else None

    @property
    def verified(self) -> bool:

        is_verified = rgx.verified.search(self.__raw_about)
        return True if is_verified else False

    @property
    def live(self) -> bool:

        check = rgx.live.findall(_get_channel_live_data(self._url))
        return True if rgx.check_live.search(check[0]) else False

    @property
    def livestream(self) -> Live:

        raw = _get_channel_live_data(self._url)
        check = rgx.live.findall(raw)
        if check and rgx.check_live.search(check[0]):
            video_id = dup_filter(rgx.video_id.findall(raw))[0]
            return Live(video_id)

    @property
    def livestreams(self) -> Optional[List[str]]:

        raw = _get_channel_live_data(self._url)
        check = rgx.live.findall(raw)
        if check and rgx.check_live.search(check[0]):
            return dup_filter(rgx.video_id.findall(raw))

    @property
    def old_streams(self) -> Optional[Dict[str, Dict[str, Any]]]:

        raw = _get_old_streams(self._url)
        ids = dup_filter(rgx.video_id.findall(raw))
        return _VideoBulk(ids)._gen_bulk() if ids else None

    def uploads(self, limit: int = 20) -> Optional[Dict[str, Dict[str, Any]]]:

        raw = _get_uploads_data(self._url)
        videos = dup_filter(rgx.uploads.findall(raw), limit)
        return _VideoBulk(videos)._gen_bulk() if videos else None

    @property
    def latest(self) -> Optional[Video]:

        raw = _get_uploads_data(self._url)
        any_video = rgx.video_id.findall(raw)
        return Video(any_video[0]) if any_video else None

    @property
    def subscribers(self) -> Optional[str]:

        subs = rgx.subscribers.findall(self.__raw_about)
        return subs[0] if subs else None

    @property
    def views(self) -> Optional[str]:

        views = rgx.views.findall(self.__raw_about)
        return views[0].split(' ')[0] if views else None

    @property
    def created_at(self) -> Optional[str]:

        joined_on = rgx.creation.findall(self.__raw_about)
        return joined_on[0] if joined_on else None

    @property
    def country(self) -> Optional[str]:

        country = rgx.country.findall(self.__raw_about)
        return country[0] if country else None

    @property
    def custom_url(self) -> Optional[str]:

        custom_urls = rgx.custom_url.findall(self.__raw_about)
        if custom_urls and '/channel/' not in custom_urls[0]:
            return custom_urls[0]

    @property
    def description(self) -> Optional[str]:

        description = rgx.description.findall(self.__raw_about)
        return description[0].replace('\\n', '\n') if description else None

    @property
    def avatar(self) -> Optional[str]:

        av = rgx.avatar.findall(self.__raw_about)
        return av[0] if av else None

    @property
    def banner(self) -> Optional[str]:

        banner = rgx.banner.findall(self.__raw_about)
        return banner[0] if banner else None

    @property
    def playlists(self) -> Optional[Dict[str, Dict[str, Any]]]:

        raw = _get_channel_playlists(self._url)
        playlists = rgx.playlists.findall(raw)
        return _PlaylistBulk(dup_filter(playlists))._gen_bulk() if playlists else None

    @property
    def info(self) -> Optional[Dict[str, any]]:
        info = {}

        def extract(pattern):
            value = pattern.findall(self.__raw_about)
            return value[0] if value else None

        patterns = [
            rgx.name, rgx.subscribers, rgx.views, rgx.creation,
            rgx.country, rgx.custom_url, rgx.avatar, rgx.banner, rgx.id,
            rgx.verified, rgx.description
        ]

        data = _Thread.run(extract, patterns)

        if data[2]:
            views = data[2].split(' ')[0]
        else:
            views = None

        curl = data[5] if data[5] and '/channel/' not in data[5] else None

        info['id'] = data[8]
        info['name'] = data[0]
        info['subscribers'] = data[1]
        info['verified'] = True if data[9] else False
        info['views'] = views
        info['created_at'] = data[3]
        info['country'] = data[4]
        info['custom_url'] = curl
        info['avatar'] = data[6]
        info['banner'] = data[7]
        info['url'] = self.url
        info['description'] = data[10].replace('\\n', '\n')

        return info

    @property
    def video_count(self) -> Optional[str]:

        if self.__given_id.startswith('UC'):
            q_term = self.__given_id
        else:
            q_term = self.id

        count = rgx.video_count.findall(_get_video_count(q_term))
        # handling channel with single digit video count
        return count[0].replace(',', '').replace('"', '').split()[0] if count else None

    @property
    def links(self) -> List[str]:

        bad_links = rgx.links.findall(self.__raw_about)
        return ['https://' + unquote(link) for link in list(set(bad_links))] if bad_links else None

    @property
    def recent_uploaded(self) -> Optional[Video]:
        raw = _get_uploads_data(self._url)
        chunk = rgx.upload_chunk.findall(raw)
        fl_1 = [data for data in chunk if not rgx.upload_chunk_fl_1.search(data)]
        fl_2 = [data for data in fl_1 if not rgx.upload_chunk_fl_2.search(data)]
        return Video(rgx.video_id.findall(fl_2[0])[0]) if fl_2 else None

    @property
    def recent_streamed(self) -> Optional[Video]:
        raw = _get_uploads_data(self._url)
        chunk = rgx.upload_chunk.findall(raw)
        fl_1 = [data for data in chunk if rgx.upload_chunk_fl_1.search(data)]
        fl_2 = [data for data in fl_1 if not rgx.upload_chunk_fl_2.search(data)]
        return Video(rgx.video_id.findall(fl_2[0])[0]) if fl_2 else None

    @property
    def upcoming(self) -> Optional[Upcoming]:
        raw = _get_upcoming_videos(self._url)
        if rgx.upcoming_check.search(raw):
            upcoming = rgx.upcoming.findall(raw)
            return Upcoming(upcoming[0]) if upcoming else None
        return None

    @property
    def all_upcoming(self) -> Optional[List[str]]:
        raw = _get_upcoming_videos(self._url)
        if rgx.upcoming_check.search(raw):
            video_ids = rgx.upcoming.findall(raw)
            return video_ids
        return None



from ._http import (
    _get_trending_video,
    _get_trending_songs,
    _get_trending_gaming_videos,
    _get_trending_news_feeds,
    _get_trending_streams,
    _get_trending_learning_videos,
    _get_trending_sports_videos
)
from .video import Video
from .utils import dup_filter
from .videobulk import _VideoBulk
from ._rgxs import _ExtraPatterns as rgx
from typing import Optional, Dict, Any



class Extras:


    @staticmethod
    def trending() -> Optional[Video]:
        """
        :return: < video object > of #1 on trending video
        """
        data = rgx.video_id.findall(_get_trending_video())
        return Video(data[0]) if data else None

    @staticmethod
    def music() -> Optional[Dict[str, Dict[str, Any]]]:
        """
        :return: list of < video object > of trending music videos
        """
        data = rgx.video_id.findall(_get_trending_songs())
        return _VideoBulk(dup_filter(data))._gen_bulk() if data else None

    @staticmethod
    def gaming() -> Optional[_VideoBulk]:
        """
        :return: list of < video object > of trending gaming videos
        """
        data = rgx.video_id.findall(_get_trending_gaming_videos())
        return _VideoBulk(dup_filter(data)) if data else None

    @staticmethod
    def news() -> Optional[_VideoBulk]:
        """
        :return: list of < video object > of trending news videos
        """
        data = rgx.video_id.findall(_get_trending_news_feeds())
        return _VideoBulk(dup_filter(data)) if data else None

    @staticmethod
    def livestreams() -> Optional[_VideoBulk]:
        """
        :return: list of < video object > of trending livestreams
        """
        data = rgx.video_id.findall(_get_trending_streams())
        return _VideoBulk(dup_filter(data)) if data else None

    @staticmethod
    def learning() -> Optional[_VideoBulk]:
        """
        :return: list of < video object > of trending educational videos
        """
        data = rgx.video_id.findall(_get_trending_learning_videos())
        return _VideoBulk(dup_filter(data)) if data else None

    @staticmethod
    def sports() -> Optional[_VideoBulk]:
        """
        :return: list of < video object > of trending sports videos
        """
        data = rgx.video_id.findall(_get_trending_sports_videos())
        return _VideoBulk(dup_filter(data)) if data else None



from ._threads import _Thread
from ._http import _get_video_data
from ._rgxs import _VideoPatterns as rgx
from typing import List, Optional, Dict, Any


class Video:

    __HEAD = 'https://www.youtube.com/watch?v='

    def __init__(self, video_id: str):

        if 'watch?v=' in video_id:
            ids = video_id.split('=')
            self._id = ids[-1]
            self._url = video_id

        elif 'youtu.be/' in video_id:
            ids = video_id.split('/')
            self._id = ids[-1]
            self._url = self.__HEAD + self._id

        else:
            self._id = video_id
            self._url = self.__HEAD + self._id

        self.__video_data = _get_video_data(self._id)

    def __repr__(self):
        if self.id:
            return f'<Video {self.url}>'
        return f'<Invalid Video Object>'

    @property
    def url(self) -> str:
        return self._url

    @property
    def id(self) -> str:
        return self._id

    @property
    def title(self) -> Optional[str]:

        data = rgx.title.findall(self.__video_data)
        return data[0] if data else None

    @property
    def views(self) -> Optional[str]:

        data = rgx.views.findall(self.__video_data)
        return data[0][:-6] if data else None

    @property
    def likes(self) -> Optional[str]:

        data = rgx.likes.findall(self.__video_data)
        return data[0] if data else None

    @property
    def duration(self) -> Optional[float]:

        data = rgx.duration.findall(self.__video_data)
        return int(data[0]) / 1000 if data else None

    @property
    def upload_date(self) -> Optional[str]:

        data = rgx.upload_date.findall(self.__video_data)
        return data[0] if data else None

    @property
    def author(self) -> Optional[str]:

        data = rgx.author_id.findall(self.__video_data)
        return data[0] if data else None

    @property
    def description(self) -> Optional[str]:

        data = rgx.description.findall(self.__video_data)
        return data[0].replace('\\n', '\n') if data else None

    @property
    def thumbnail(self) -> Optional[str]:

        data = rgx.thumbnail.findall(self.__video_data)
        return data[0] if data else None

    @property
    def tags(self) -> Optional[List[str]]:

        data = rgx.tags.findall(self.__video_data)
        return data[0].split(',') if data else None

    @property
    def streamed(self) -> bool:
        if rgx.is_streamed.search(self.__video_data):
            return True
        return False

    @property
    def premiered(self) -> bool:
        if rgx.is_premiered.search(self.__video_data):
            return True
        return False

    @property
    def info(self) -> Dict[str, Any]:
        info = {}

        def _get_data(pattern):
            value = pattern.findall(self.__video_data)
            return value[0] if value else None

        patterns = [
            rgx.title, rgx.views, rgx.likes, rgx.duration, rgx.author_id,
            rgx.upload_date, rgx.thumbnail, rgx.tags, rgx.description,
            rgx.is_streamed, rgx.is_premiered
        ]

        data = _Thread.run(_get_data, patterns)

        info['title'] = data[0]
        info['id'] = self._id
        info['views'] = data[1][:-6] if data[1] else None
        info['likes'] = data[2]
        info['streamed'] = data[9] if data[9] else None
        info['premiered'] = data[10] if data[10] else None
        info['duration'] = int(data[3]) / 1000 if data[3] else None
        info['author'] = data[4]
        info['upload_date'] = data[5]
        info['url'] = self._url
        info['thumbnail'] = data[6]
        info['tags'] = data[7].split(',') if data[7] else None
        info['description'] = data[8].replace('\\n', '\n') if data[8] else None

        return info



import re


class _ChannelPatterns:
    name = re.compile('channelMetadataRenderer\":{\"title\":\"(.*?)\"')
    id = re.compile('channelId\":\"(.*?)\"')
    verified = re.compile('"label":"Verified"')
    check_live = re.compile('{"text":"LIVE"}')
    live = re.compile("thumbnailOverlays\":\[(.*?)]")
    video_id = re.compile('videoId\":\"(.*?)\"')
    uploads = re.compile("gridVideoRenderer\":{\"videoId\":\"(.*?)\"")
    subscribers = re.compile("}},\"simpleText\":\"(.*?) ")
    views = re.compile("viewCountText\":{\"simpleText\":\"(.*?)\"}")
    creation = re.compile("{\"text\":\"Joined \"},{\"text\":\"(.*?)\"}")
    country = re.compile("country\":{\"simpleText\":\"(.*?)\"}")
    custom_url = re.compile("canonicalChannelUrl\":\"(.*?)\"")
    description = re.compile("{\"description\":{\"simpleText\":\"(.*?)\"}")
    avatar = re.compile("height\":88},{\"url\":\"(.*?)\"")
    banner = re.compile("width\":1280,\"height\":351},{\"url\":\"(.*?)\"")
    playlists = re.compile("{\"url\":\"/playlist\?list=(.*?)\"")
    video_count = re.compile("videoCountText\":{\"runs\":\[{\"text\":(.*?)}")
    links = re.compile("q=https%3A%2F%2F(.*?)\"")
    upload_chunk = re.compile("gridVideoRenderer\":{(.*?)\"navigationEndpoint")
    upload_chunk_fl_1 = re.compile("simpleText\":\"Streamed")
    upload_chunk_fl_2 = re.compile("default_live.")
    upcoming_check = re.compile("\"title\":\"Upcoming live streams\"")
    upcoming = re.compile("gridVideoRenderer\":{\"videoId\":\"(.*?)\"")


class _VideoPatterns:
    video_id = re.compile('videoId\":\"(.*?)\"')
    title = re.compile("title\":\"(.*?)\"")
    duration = re.compile("approxDurationMs\":\"(.*?)\"")
    upload_date = re.compile("uploadDate\":\"(.*?)\"")
    author_id = re.compile("channelIds\":\[\"(.*?)\"")
    description = re.compile("shortDescription\":\"(.*)\",\"isCrawlable")
    tags = re.compile("<meta name=\"keywords\" content=\"(.*?)\">")
    is_streamed = re.compile("simpleText\":\"Streamed live")
    is_premiered = re.compile("dateText\":{\"simpleText\":\"Premiered")
    views = re.compile("videoViewCountRenderer\":{\"viewCount\":{\"simpleText\":\"(.*?)\"")
    likes = re.compile("toggledText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) ")
    thumbnail = re.compile("playerMicroformatRenderer\":{\"thumbnail\":{\"thumbnails\":\[{\"url\":\"(.*?)\"")


class _PlaylistPatterns:
    name = re.compile("{\"title\":\"(.*?)\"")
    video_count = re.compile("stats\":\[{\"runs\":\[{\"text\":\"(.*?)\"")
    video_id = re.compile("videoId\":\"(.*?)\"")
    thumbnail = re.compile("og:image\" content=\"(.*?)\?")


class _ExtraPatterns:
    video_id = re.compile("videoId\":\"(.*?)\"")


class _QueryPatterns:
    channel_id = re.compile("channelId\":\"(.*?)\"")
    video_id = re.compile("videoId\":\"(.*?)\"")
    playlist_id = re.compile("playlistId\":\"(.*?)\"")



from ._http import _get_video_data
from ._threads import _Thread
from ._rgxs import _VideoPatterns as rgx
from typing import List, Optional, Dict, Any


class _VideoBulk:

    def __init__(self, iterable: list):
        self._video_ids = iterable
        self.__source_data = self.__fetch_all()

    def __fetch_all(self):

        def fetch_bulk_source(video_id):
            return _get_video_data(video_id)

        return _Thread.run(fetch_bulk_source, self._video_ids)

    @staticmethod
    def _gen_info(source: str) -> Dict[str, Any]:
        info = {}

        def _get_data(pattern):
            d = pattern.findall(source)
            return d[0] if d else None

        patterns = [
            rgx.title, rgx.views, rgx.likes, rgx.duration, rgx.author_id,
            rgx.upload_date, rgx.thumbnail, rgx.tags, rgx.description,
            rgx.is_streamed, rgx.is_premiered, rgx.video_id
        ]

        data = _Thread.run(_get_data, patterns)

        info['id'] = data[11]
        info['title'] = data[0]
        info['views'] = data[1]
        info['likes'] = data[2]
        info['duration'] = int(data[3]) / 1000 if data[3] else None
        info['author'] = data[4]
        info['upload_date'] = data[5]
        info['streamed'] = True if data[9] else False
        info['premiered'] = True if data[10] else False
        info['thumbnail'] = data[6]
        info['tags'] = data[7].split(',') if data[7] else None
        info['description'] = data[8].replace('\\n', '\n') if data[8] else None
        info['url'] = f'https://www.youtube.com/watch?v={data[11]}'

        return info

    def _gen_bulk(self) -> Dict[str, Dict[str, Any]]:
        info_list = [self._gen_info(source) for source in self.__source_data]
        return {info.pop('id'): info for info in info_list}



from .video import Video
from ._http import _get_video_data
from typing import Optional, Dict, Any
from ._rgxs import _VideoPatterns as rgx


class Upcoming(Video):

    def __init__(self, video_id: str):
        super().__init__(video_id)

    def duration(self) -> float:
        return 0.0

    def views(self) -> Optional[str]:
        return None

    @property
    def info(self) -> Dict[str, Any]:
        """
        :return: dict containing the whole info of the video
        """
        raw = _get_video_data(self._id)

        def _get_data(pattern):
            data = pattern.findall(raw)
            return data[0] if len(data) > 0 else None

        patterns = [
            rgx.title, rgx.likes, rgx.author_id,
            rgx.upload_date, rgx.thumbnail, rgx.tags, rgx.description
        ]

        data = _Thread.run(_get_data, patterns)

        description = data[6].replace('\\n', '\n') if data[6] else None

        return {
            'title': data[0],
            'id': self._id,
            'views': None,
            'likes': data[1],
            'duration': None,
            'author': data[2],
            'upload_date': data[3],
            'url': self._url,
            'thumbnail': data[4],
            'tags': data[5].split(','),
            'description': description,
        }



from .utils import _fetch_webpage, parser


def _get_channel_about(head: str) -> str:
    url = head + '/about'
    return _fetch_webpage(url)


def _get_channel_live_data(head: str) -> str:
    url = head + '/videos?view=2&live_view=501'
    return _fetch_webpage(url)


def _get_old_streams(head: str) -> str:
    url = head + '/videos?view=2&live_view=503'
    return _fetch_webpage(url)


def _get_uploads_data(head: str) -> str:
    url = head + '/videos?view=0&sort=dd&flow=grid'
    return _fetch_webpage(url)


def _get_channel_playlists(head: str) -> str:
    url = head + '/playlists'
    return _fetch_webpage(url)


def _get_upcoming_videos(head: str) -> str:
    url = head + '/videos?view=2&live_view=502'
    return _fetch_webpage(url)


def _get_video_count(channel_id: str) -> str:
    head = 'https://www.youtube.com/results?search_query='
    tail = '&sp=EgIQAg%253D%253D'
    url = head + channel_id + tail
    return _fetch_webpage(url)


def _get_video_data(video_id: str) -> str:
    url = f'https://www.youtube.com/watch?v={video_id}'
    return _fetch_webpage(url)


def _get_playlist_data(playlist_id: str) -> str:
    url = 'https://www.youtube.com/playlist?list=' + playlist_id
    return _fetch_webpage(url)


def _get_trending_video() -> str:
    return _fetch_webpage('https://www.youtube.com/feed/trending')


def _get_trending_songs() -> str:
    return _fetch_webpage('https://www.youtube.com/feed/music')


def _get_trending_gaming_videos() -> str:
    return _fetch_webpage('https://www.youtube.com/gaming')


def _get_trending_news_feeds() -> str:
    return _fetch_webpage('https://www.youtube.com/news')


def _get_trending_streams() -> str:
    return _fetch_webpage('https://www.youtube.com/live')


def _get_trending_learning_videos() -> str:
    return _fetch_webpage('https://www.youtube.com/learning')


def _get_trending_sports_videos() -> str:
    return _fetch_webpage('https://www.youtube.com/sports')


def _find_videos(query: str) -> str:
    head = 'https://www.youtube.com/results?search_query='
    tail = '&sp=EgIQAQ%253D%253D'
    parsed_query = parser(query)
    return _fetch_webpage(head + parsed_query + tail)


def _find_channels(query: str) -> str:
    head = 'https://www.youtube.com/results?search_query='
    tail = '&sp=EgIQAg%253D%253D'
    parsed_query = parser(query)
    return _fetch_webpage(head + parsed_query + tail)


def _find_playlists(query: str) -> str:
    head = 'https://www.youtube.com/results?search_query='
    tail = '&sp=EgIQAw%253D%253D'
    parsed_query = parser(query)
    return _fetch_webpage(head + parsed_query + tail)



import concurrent.futures


class _Thread:

    def __init__(self):
        pass

    @staticmethod
    def run(func, args: list) -> list:
        num = len(args) if len(args) > 0 else 1
        if args:
            with concurrent.futures.ThreadPoolExecutor(max_workers=num) as exe:
                return list(exe.map(func, args))



from ._threads import _Thread
from .utils import dup_filter
from ._http import _get_playlist_data
from ._rgxs import _PlaylistPatterns as rgx
from typing import List, Dict, Any


class _PlaylistBulk:

    def __init__(self, iterable: list):
        self._playlist_ids = iterable
        self.__bulk = self.__fetch_all()

    def __fetch_all(self):

        def fetch_bulk_source(playlist_id):
            return _get_playlist_data(playlist_id)

        return _Thread.run(fetch_bulk_source, self._playlist_ids)

    @staticmethod
    def _get_info(source: str) -> Dict[str, Any]:
        info = {}

        def _get_data(pattern):
            d = pattern.findall(source)
            return d[0] if d else None

        patterns = [rgx.name, rgx.video_count, rgx.thumbnail]

        data = _Thread.run(_get_data, patterns)

        info['name'] = data[0]
        info['video_count'] = data[1]
        info['thumbnail'] = data[2]
        info['url'] = None
        info['videos'] = dup_filter(rgx.video_id.findall(source))

        return info

    def _gen_bulk(self) -> Dict[str, Dict[str, Any]]:
        info_list = []
        for pl_data in self.__bulk:
            info = self._get_info(pl_data)
            for pl_id in self._playlist_ids:
                if pl_id in pl_data:
                    info['id'] = pl_id
                    info['url'] = f'https://www.youtube.com/playlist?list={pl_id}'
            info_list.append(info)

        return {info.pop('id'): info for info in info_list}



class TooManyRequests(Exception):
    """
    Raised when the user has sent too many requests to the API.
    """
    def __init__(self, message):
        self.message = message


class InvalidURL(Exception):
    """
    Raised when the URL is invalid.
    """
    def __init__(self, message):
        self.message = message


class BadURL(Exception):
    """
    Raised when the URL format is invalid.
    """
    def __init__(self, message):
        self.message = message


class AIOError(Exception):
    """
    Raised when the response is not what we expected.
    """
    def __init__(self, message):
        self.message = message



from .video import Video
from ._threads import _Thread
from ._http import _get_video_data
from ._rgxs import _VideoPatterns as rgx


class Live(Video):

    def __init__(self, video_id: str):
        super().__init__(video_id)

    @property
    def duration(self) -> float:
        return 0.0

    @property
    def streamed(self) -> bool:
        return True

    @property
    def info(self) -> dict:
        """
        :return: dict containing the whole info of the video
        """
        raw = _get_video_data(self._id)

        def _get_data(pattern):
            data = pattern.findall(raw)
            return data[0] if len(data) > 0 else None

        patterns = [
            rgx.title, rgx.views, rgx.likes, rgx.duration, rgx.author_id,
            rgx.upload_date, rgx.thumbnail, rgx.tags, rgx.description
        ]

        data = _Thread.run(_get_data, patterns)

        return {
            'title': data[0],
            'id': self._id,
            'views': data[1][:-6] if data[1] else None,
            'likes': data[2],
            'duration': self.duration,
            'author': data[4],
            'upload_date': data[5],
            'url': self._url,
            'thumbnail': data[6],
            'tags': data[7].split(','),
            'description': data[8].replace('\\n', '\n') if data[8] else None,
            'streamed': self.streamed,
        }



from ._threads import _Thread
from ._http import _get_channel_about
from ._rgxs import _ChannelPatterns as rgx
from typing import List, Optional, Dict, Any


class _ChannelBulk:

    __HEAD = 'https://www.youtube.com/channel/'

    def __init__(self, iterable: list):
        self._channel_ids = iterable
        self.__bulk_data = self.__fetch_all

    @property
    def __fetch_all(self):
        urls = [self.__HEAD + id for id in self._channel_ids]
        return _Thread.run(_get_channel_about, urls)

    def _gen_bulk(self):
        info_list = [self._gen_info(source) for source in self.__bulk_data]
        return {info.pop('id'): info for info in info_list}

    @staticmethod
    def _gen_info(source: str) -> Optional[Dict[str, Dict[str, Any]]]:
        info = {}

        def extract(pattern):
            d = pattern.findall(source)
            return d[0] if d else None

        patterns = [
            rgx.name, rgx.subscribers, rgx.views, rgx.creation,
            rgx.country, rgx.custom_url, rgx.avatar, rgx.banner, rgx.id,
            rgx.verified, rgx.description
        ]

        data = _Thread.run(extract, patterns)

        if data[2]:
            views = data[2].split(' ')[0]
        else:
            views = None

        if data[10]:
            description = data[10].replace('\\n', '\n')
        else:
            description = None

        curl = data[5] if data[5] and '/channel/' not in data[5] else None

        info['id'] = data[8]
        info['name'] = data[0]
        info['subscribers'] = data[1]
        info['views'] = views
        info['created_at'] = data[3]
        info['country'] = data[4]
        info['custom_url'] = curl
        info['avatar'] = data[6]
        info['banner'] = data[7]
        info['url'] = f'https://www.youtube.com/channel/{data[8]}'
        info['description'] = description.replace('\\n', '\n') if description else None
        info['verified'] = True if data[9] else False

        return info



import pathlib
from setuptools import setup, find_packages


here = pathlib.Path(__file__).parent.resolve()

with open('README.rst') as f:
    readme = f.read()


setup(
    name='aiotube',
    version='1.5.7',
    description='Access YouTube Public Data without YouTube API',
    long_description=readme,
    long_description_content_type="text/x-rst",
    url='https://github.com/jnsougata/aiotube',
    author='jnsougata',
    author_email='jnsougata@gmail.com',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    keywords='youtube, youtube-data, youtube-api, youtube-data-api-v3',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.8.0',
    install_requires=['urllib3'],
    project_urls={
        'Documentation': 'https://aiotube.readthedocs.io/en/latest/',
        'Source': 'https://github.com/jnsougata/aiotube'
    },
)



from .utils import dup_filter
from .video import Video
from .channel import Channel
from .playlist import Playlist
from .videobulk import _VideoBulk
from .channelbulk import _ChannelBulk
from .playlistbulk import _PlaylistBulk
from ._rgxs import _QueryPatterns as rgx
from typing import Optional, Optional, Dict, Any
from ._http import _find_videos, _find_channels, _find_playlists


class Search:

    @staticmethod
    def video(keywords: str) -> Optional[Video]:
        video_ids = rgx.video_id.findall(_find_videos(keywords))
        return Video(video_ids[0]) if video_ids else None

    @staticmethod
    def channel(keywords: str) -> Optional[Channel]:
        channel_ids = rgx.channel_id.findall(_find_channels(keywords))
        return Channel(channel_ids[0]) if channel_ids else None

    @staticmethod
    def playlist(keywords: str) -> Optional[Playlist]:
        playlist_ids = rgx.playlist_id.findall(_find_playlists(keywords))
        return Playlist(found[0]) if found else None

    @staticmethod
    def videos(keywords: str, limit: int = 20) -> Optional[Dict[str, Dict[str, Any]]]:
        video_ids = rgx.video_id.findall(_find_videos(keywords))
        pure_list = dup_filter(limit=limit, iterable=video_ids)
        return _VideoBulk(pure_list)._gen_bulk() if pure_list else None

    @staticmethod
    def channels(keywords: str, limit: int = 20) -> Optional[Dict[str, Dict[str, Any]]]:
        channel_ids = rgx.channel_id.findall(_find_channels(keywords))
        pure_list = dup_filter(limit=limit, iterable=channel_ids)
        return _ChannelBulk(pure_list)._gen_bulk() if pure_list else None

    @staticmethod
    def playlists(keywords: str, limit: int = 20) -> Optional[Dict[str, Dict[str, Any]]]:
        playlist_ids = rgx.playlist_id.findall(_find_playlists(keywords))
        pure_list = dup_filter(limit=limit, iterable=playlist_ids)
        return _PlaylistBulk(pure_list)._gen_bulk() if pure_list else None



from urllib.request import urlopen
from collections import OrderedDict
from urllib.error import HTTPError, URLError
from .errors import TooManyRequests, InvalidURL, BadURL, AIOError


__all__ = ['dup_filter', 'parser']


def _fetch_webpage(url: str):

    try:
        return urlopen(url).read().decode()
    except HTTPError as e:
        if e.code == 404:
            raise InvalidURL('can not find anything with the requested url')
        if e.code == 429:
            raise TooManyRequests('you are being rate-limited for sending too many requests')
    except URLError:
        raise BadURL('url does not match any supported format')
    except Exception as e:
        raise AIOError(f'{e!r}')


def dup_filter(iterable: list, limit: int = None) -> list:
    if iterable:
        lim = limit if limit else len(iterable)
        converted = list(OrderedDict.fromkeys(iterable))
        if len(converted) - lim > 0:
            return converted[:-len(converted) + lim]
        else:
            return converted
    return []


def parser(kw: str):
    return kw.replace(" ", '+')



from ._threads import _Thread
from .utils import dup_filter
from .videobulk import _VideoBulk
from ._http import _get_playlist_data
from ._rgxs import _PlaylistPatterns as rgx
from typing import List, Optional, Dict, Any


class Playlist:

    def __init__(self, playlist_id: str):
        if 'youtube.com' in playlist_id:
            self.id = playlist_id.split('list=')[-1]
        else:
            self.id = playlist_id

        self.__playlist_data = _get_playlist_data(self.id)

    def __repr__(self):
        return f'<Playlist {self.url}>'

    @property
    def name(self) -> Optional[str]:
        names = rgx.name.findall(self.__playlist_data)
        return names[0] if names else None

    @property
    def url(self) -> Optional[str]:
        return f'https://www.youtube.com/playlist?list={self.id}'

    @property
    def video_count(self) -> Optional[str]:
        video_count = rgx.video_count.findall(self.__playlist_data)
        return video_count[0] if video_count else None

    @property
    def videos(self) -> _VideoBulk:
        videos = rgx.video_id.findall(self.__playlist_data)
        return _VideoBulk(dup_filter(iterable=videos))

    @property
    def thumbnail(self) -> Optional[str]:
        thumbnails = rgx.thumbnail.findall(self.__playlist_data)
        return thumbnails[0] if thumbnails else None
    
    @property
    def info(self) -> Dict[str, Any]:
        info = {}

        def _get_data(pattern):
            d = pattern.findall(self.__playlist_data)
            return d[0] if d else None

        patterns = [rgx.name, rgx.video_count, rgx.thumbnail]

        data = _Thread.run(_get_data, patterns)

        info['id'] = self.id
        info['name'] = data[0]
        info['video_count'] = data[1]
        info['thumbnail'] = data[2]
        info['url'] = self.url
        info['videos'] = dup_filter(rgx.video_id.findall(raw))

        return info
        